// kvector.cpp : Simple implementation of STL vector

#include "stdafx.h"
#include <iostream>
#include <ctype.h>
#include <algorithm>
using namespace std;

//! \file kvector.cpp
//
/*! \mainpage kvector documentation page
 * This provides a way to study the kvector code
 * by following the links in several ways.
 * Let me know if it is useful.
 */

const int DEFSIZE=5; //!< default vector size

const int INCR=5; //!< how much to increase the array size by 

//! The kvector class
/*! We are making a vector class. A vector is like an array but the
    length can vary. We will use templates so  that it can work
    with an array of any type and make it longer as needed.
    This is similar to the way the Vector class from the STL works,
    but much simpler.
    I have also added some other methods and operators to better support
    the Vector data type.
*/ 
template <class MYTYPE>
class kvector {
	public:
		kvector(void); //!< default constructor
		kvector(int);  //!< set initial vector size to size
		~kvector(void) { if(vptr != NULL) delete [] vptr;}

		void add(MYTYPE x); // add data value x to next open slot
		void show(void); // print the current contents of the vector
		int get_length(void) { return length;}

		 //! access
		 /*! the access function can be used on either side of the =
		    since it returns a reference to the array member being accessed.
		*/ 
		MYTYPE & access(int index) 
		{
			if((index >= 0) && (index < length))
				return vptr[index];
			else
				throw "Bad index";
		}

		//! overloaded array access operator
		/*! makes our vector look like an array since we can access the elements
		   using the [] operator. It also checks that we are not trying to access
		   something we shouldn't.
		   Note that it is very similar to the access() method 
		*/ 
		MYTYPE & operator[](int index) 
		{
			if((index >= 0) && (index < length))
				return vptr[index];
			else
				throw "Bad index";
		}

		//! overloaded output operator
		/*! prints the contents in a line. Assumes the base type (MYTYPE)
		   has a << operator defined.
		*/ 
		friend ostream & operator<<(ostream & os, kvector<MYTYPE> & k)
		{
			for(int i=0; i < k.get_length(); i++) 
	   				os << k[i] << " ";
			os << endl;
			return os;
		}
		
		//! internal public iterator class
		/*! now we define the internal iterator class. This allows us 
		   to create pointer like objects that will work on our Vector
		   without the user having to know that there is an array inside.
		 I added the begin() and end() methods to the kvector class to support this.
		 The essential feature of an iterator class is that it overload the pointer operators
		 like * and movement operators like ++ and --.
		*/ 
		class iterator {
		public:
   			iterator() {ptr = 0;}
   			iterator(MYTYPE * m) {ptr = m;}
			iterator(iterator & i) {ptr = i.get_ptr();}  //!< copy constructor
			MYTYPE * get_ptr() { return ptr;}

			// since the pointer hidden inside the iterator is an regular
			// array pointer, then iterator++ turns into pointer++.
			// If this had been a list object, then the iterator++ would have
			// turned into the pointer = pointer->next kind of structure.
			// We also need one for prefix increment for copy() to work.
   			void operator++(int) { ptr++; }   //!< pointer post-increment
   			void operator++() { ++ptr; }   //!< pointer pre-increment

			// probably the bitwise copy of the default assignment operator
			// would have worked, but I wanted to be sure.
   			void operator=(iterator p) { ptr = p.get_ptr(); }  //!< assignment operator

			// again, since the iterator hides a regular array pointer, the
			// dereference operator becomes the pointer dereference
   			MYTYPE & operator*(void) { return *ptr;}  //!< dereference pointer

			// comparison operators since the default might work but we want to be sure.
   			bool operator==(iterator i) { return (this->ptr == i.get_ptr());}
   			bool operator!=(iterator k) { return (this->ptr != k.get_ptr());}
		private:
   			MYTYPE * ptr;  //!< the actual pointer to the internal array in the vector
        };

		//! returns initial iterator for the kvector
		/*! this creates a new iterator object and initializes it to
		   the beginning of the underlying array (vptr).
		   It then returns the iterator.
		   Note that x is a pointer because new returns a pointer,
		   but we have to return the actual iterator object because that is
		   what is expected of begin().
		*/
		iterator begin() {
				iterator  *x = new iterator(vptr);
				return *x;
		}

		//! returns end iterator for the kvector
		/*! This also creates a new iterator and points it
		   at the memory just after the end of the array.
		  it does this by getting the address of the last cell in the array
		  &vptr[length-1] is the address of the last element in the array
		  Then we add 1 to get the address of the next place after the end
		  of the array.
		*/
		iterator end() { 
				iterator  *x = new iterator( (MYTYPE *)(&vptr[length-1]) + 1);
				return *x;
		}

	protected:
		MYTYPE * vptr;  //!< pointer to current vector array
		int count;   //!< current number of used slots in vector
		int length;  //!< current number of slots in vector
}; 

//! create the initial vector of length DEFSIZE
template <class MYTYPE>
kvector<MYTYPE>::kvector()
{
	length = DEFSIZE;
	count = 0;
	if( (vptr = new MYTYPE[length]) == NULL) {
		cerr << "No memory for vector of length " << length << endl;
	}
} // constructor

//! create the initial vector of length size
template <class MYTYPE>
kvector<MYTYPE>::kvector(int size)
{
	length = size;
	count = 0;
	if( (vptr = new MYTYPE[length]) == NULL) {
		cerr << "No memory for vector of length " << length << endl;
	}
} // 1 arg constructor

//! print the current data elements in the vector
template <class MYTYPE>
void
kvector<MYTYPE>::show(void)
{
	cout << "vector = (";
	for(int i = 0; i < count; i++) {
		cout << vptr[i];
		if(i+1 != count) cout << ","; // don't print trailing comma
	}
	cout << ")" << endl;
} // show

/*! adds the integer x to the next slot in the vector
   if the vector is full, it allocates a new array, copies the values,
   and adjusts the pointers
*/
template <class MYTYPE>
void
kvector<MYTYPE>::add(MYTYPE x)
{
	/*! count is the number of used slots
	    length is the number of slots
		if the count and length are equal, then the array
		is full and we need to allocate a new bigger array.
		We make the array bigger in chunks to avoid having to run this code
		too often.
		after the new array is created, we copy the old array elements 
		to the new array.
		Changing the internal kvector array pointer (vptr) to
		point to the new array completes the work.
		We can delete the old array now.
		length is changed to the new array size.
		Whether or not the array was extended, we add the
		parameter x to the array.
	*/
	if( count == length) {  // make new array
		cout << "making new array" << endl;
		MYTYPE * tptr = new MYTYPE[length + INCR];
		if(tptr == NULL) { // this is fatal
			cerr << "No memory for new array" << endl;
			throw ("no memory");
		}
		// copy the old array elements to the new array.
		for(int i=0; i < count; i++) tptr[i] = vptr[i];
//		delete vptr;  // we don't need this anymore
		vptr = tptr;  // make the new array the current array
		length += INCR;
	}
	vptr[count++] = x;
} // add

//! run test code on integers
void
vector_int()
{
	kvector<int> v;

	try {
		// fill up the vector
		cout << "putting in initial data" << endl;
		for(int i=0; i < v.get_length(); i++) v.add(i);
		v.show();
		cout << "now adding 5 more" << endl;
		for(int j=0; j < 5; j++) v.add(j);
		v.show();

		cout << "print using the overloaded [] operator" << endl;
		for(int k=0; k < v.get_length(); k++) cout << v[k] << " ";
		cout << endl;

		/* since the [] operator returns a reference, we can treat it like a variable
		   and use it on the left side of an asignment
	   */
		cout << " changing element 0 to 99 using []" << endl;
		v[0] = 99;
		cout << v;

		// try the access method on either side of the =
		cout << "element 1 is " << v.access(1) << endl;
		cout << " changing element 1 to 42 using access()" << endl;
		v.access(1) = 42;
		cout << v;
		cout << "testing the array boundary exception" << endl;
		v[100] = 87;
	}
	catch(char * str ) {
			cerr << str << endl;
	}
	catch(...) {
			cerr << "Some unknown exception caught" << endl;
	}

	/*
	   now test out the iterator class
	   since the iterator class is inside the kvector class,
	   we have to use the scope resolution operator (::) to use it.
	   We will also test the iterator by using the copy algorithm
	   from the STL.
   */
	kvector<int>::iterator ptr;
	try {
			cout << "printing the vector using iterator" << endl;
			ptr = v.begin();
			while(ptr != v.end()) {
				cout << *ptr << " ";
				ptr++;
			}
			cout << endl;
			cout << "printing the vector using copy()" << endl;
			copy(v.begin(),v.end(),ostream_iterator<int>(cout," "));
			cout << endl;
	}
	catch(char * str ) {
			cerr << str << endl;
	}
	catch(...) {
			cerr << "Some unknown exception caught" << endl;
	}

} // vector_int

//! run test code on floats
void
vector_float()
{
	kvector<double> v;

	try {
		// fill up the vector
		cout << "putting in initial data" << endl;
		for(int i=0; i < v.get_length(); i++) v.add(i/3.0);
		v.show();
		cout << "now adding 5 more" << endl;
		for(int j=0; j < 5; j++) v.add(j/3.0);
		v.show();

		cout << "print using the overloaded [] operator" << endl;
		for(int k=0; k < v.get_length(); k++) cout << v[k] << " ";
		cout << endl;

		/*
		   since the [] operator returns a reference, we can treat it like a variable
		   and use it on the left side of an asignment
	   */
		cout << " changing element 0 to 99.4 using []" << endl;
		v[0] = 99.4;
		cout << v;

		// try the access method on either side of the =
		cout << "element 1 is " << v.access(1) << endl;
		cout << " changing element 1 to 42.4 using access()" << endl;
		v.access(1) = 42.4;
		cout << v;
		cout << "testing the array boundary exception" << endl;
		v[100] = 87.4;
	}
	catch(char * str ) {
			cerr << str << endl;
	}
	catch(...) {
			cerr << "Some unknown exception caught" << endl;
	}

	/*
	   now test out the iterator class
	   since the iterator class is inside the kvector class,
	   we have to use the scope resolution operator (::) to use it.
	   We will also test the iterator by using the copy algorithm
	   from the STL.
   */
	kvector<double>::iterator ptr;
	try {
			cout << "printing the vector using iterator" << endl;
			ptr = v.begin();
			while(ptr != v.end()) {
				cout << *ptr << " ";
				ptr++;
			}
			cout << endl;
			cout << "printing the vector using copy()" << endl;
			copy(v.begin(),v.end(),ostream_iterator<double>(cout," "));
			cout << endl;
	}
	catch(char * str ) {
			cerr << str << endl;
	}
	catch(...) {
			cerr << "Some unknown exception caught" << endl;
	}

} // vector_float


int main(int argc, char* argv[])
{
	char ch;
	cout << "Enter i for integer test or f for float test" << endl;
	cin >> ch;
	cin.ignore();
	if(tolower(ch) == 'i')
		vector_int();
	else
		vector_float();
	return 0;
}
